/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1996  Hans Verkuil
 *  Copyright (C) 2015  Harry Sintonen.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"

#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

extern int read();

caddr_t mmap64 (caddr_t addr, size_t len, int prot, int flags, int fd, off64_t offset)
{
  struct mmap_mem *m;
  usetup;

  if (flags & MAP_FIXED)
    {
      errno = ENOMEM;
      return MAP_FAILED;
    }
  if (!(flags & MAP_ANON) && (fd < 0 || fd >= NOFILE || !u.u_ofile[fd]))
    {
      errno = EBADF;
      return MAP_FAILED;
    }
  m = malloc(sizeof(struct mmap_mem));
  if (m == NULL)
    {
      errno = ENOMEM;
      return MAP_FAILED;
    }
  m->addr = valloc(len);
  if (m->addr == NULL)
    {
      free(m);
      errno = ENOMEM;
      return MAP_FAILED;
    }
  if (!(flags & MAP_ANON))
    {
      off64_t curoff = lseek64(fd, 0LL, SEEK_CUR);

      lseek64(fd, offset, SEEK_SET);
      read(fd, m->addr, len);
      lseek64(fd, curoff, SEEK_SET);
    }
  m->length = len;
  m->prot = prot;
  m->fd = fd;
  m->flags = flags;
  m->offset = offset;
  m->next = u.u_mmap;
  u.u_mmap = m;
  return m->addr;
}
