/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: pooled-alloc.c,v 1.1 2012/02/15 16:56:35 morgoth Exp $
 *
 */

#define _KERNEL
#include "ixemul.h"
#include "ixalloc.h"
#include "kprintf.h"

#include <exec/memory.h>
#include <stddef.h>

#ifndef USE_OLD_ALLOC

/* This is not realy used for now. We need it only to init/cleanup
   pools. */

/* NOTE: currently only two pools are supported, MEMF_PUBLIC and
	 ! MEMF_PUBLIC. No MEMF_CHIP pools are needed by the library
	 and are thus not supported */

void *ix_prv_mempool; /* Private pool */
void *ix_pub_mempool; /* Public pool */

void
init_buddy (void)
{
  /* have to differentiate between PUBLIC and PRIVATE memory here, sigh.
     PRIVATE memory can safely be accessed by using a semaphore, PUBLIC
     memory however is allocated and free'd inside Forbid(), and using a
     semaphore there would possibly break a Forbid..
     Note: this is safe for use in GigaMem, as GigaMem only uses non-PUBLIC
     memory, if you don't fiddle with attribute masks.. */

  ix_prv_mempool = CreatePool(MEMF_SWAP|MEMF_SEM_PROTECTED, IXALLOC_PRV_POOL_SIZE, IXALLOC_PRV_POOL_SIZE / 2); /* Private */
  ix_pub_mempool = CreatePool(MEMF_PUBLIC, IXALLOC_PUB_POOL_SIZE, IXALLOC_PUB_POOL_SIZE / 2); /* Public */

  if(ix_prv_mempool == NULL || ix_pub_mempool == NULL)
    {
      ix_panic ("buddy-system: Failed to create memory pools!");
      Wait (0);
    }
}

void cleanup_buddy(void)
{
  if(ix_pub_mempool)
    DeletePool(ix_pub_mempool);

  if(ix_prv_mempool)
    DeletePool(ix_prv_mempool);
}

#endif
