/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * Copyright (C) 2021 Harry Sintonen
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * C library -- setjmp, longjmp
 *
 *      longjmp(a,v)
 * will generate a "return(v)" from
 * the last call to
 *      setjmp(a)
 * by restoring registers from the stack,
 * and a struct sigcontext, see <signal.h>
 */

#define _KERNEL
#include "ixemul.h"
#include "defs.h"
#include "ix_internals.h"

#include <sys/syscall.h>

#ifdef NATIVE_MORPHOS

#if _JBLEN != 146
#error "_JBLEN is no longer compatile. This code likely needs to be adjusted!"
#endif

/* The __obsolete_XXX_setjmp functions are called by 68k
 * programs compiled with old 68k _JBLEN.
 * Allocate a bigger buffer, store a pointer to it in the
 * 68k buffer, and set the 'cr' field to 0, so that longjmp
 * can recognize that it is given a 68k buffer.
 */

int __obsolete_sigsetjmp();
int __obsolete__setjmp();
int __obsolete_setjmp();

static int _trampoline___obsolete_sigsetjmp();
static int _trampoline___obsolete__setjmp();
static int _trampoline___obsolete_setjmp();

const struct EmulLibEntry _gate___obsolete_sigsetjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline___obsolete_sigsetjmp
};

const struct EmulLibEntry _gate___obsolete__setjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline___obsolete__setjmp
};

const struct EmulLibEntry _gate___obsolete_setjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline___obsolete_setjmp
};


static int _trampoline_sigsetjmp();
static int _trampoline__setjmp();
static int _trampoline_setjmp();
static int _trampoline_siglongjmp();
static int _trampoline__longjmp();
static int _trampoline_longjmp();

const struct EmulLibEntry _gate_sigsetjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline_sigsetjmp
};

const struct EmulLibEntry _gate__setjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline__setjmp
};

const struct EmulLibEntry _gate_setjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline_setjmp
};

const struct EmulLibEntry _gate_siglongjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline_siglongjmp
};

const struct EmulLibEntry _gate__longjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline__longjmp
};

const struct EmulLibEntry _gate_longjmp = {
  TRAP_LIBRESTORE, 0, (void(*)())_trampoline_longjmp
};


asm(" .section \".text\"");

ENTRY(_trampoline___obsolete_sigsetjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	lwz     4,8(5)
	b       __obsolete_sigsetjmp
");

ENTRY(__obsolete_sigsetjmp)
asm("
	cmpi    0,4,0
	stw     4,0(3)
	addi    3,3,4
	bne     __obsolete_setjmp
	b       __obsolete__setjmp
");


ENTRY(_trampoline___obsolete__setjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	b       __obsolete__setjmp
");

ENTRY(__obsolete__setjmp)
asm("
	li      6,0             /* dummy signal mask */
	li      7,0             /*  and onstack */
	b       __obsolete_setup_setjmp
");


ENTRY(_trampoline___obsolete_setjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	b       __obsolete_setjmp
");

ENTRY(__obsolete_setjmp)
asm("
__obsolete_setjmp:
	stwu    1,-16(1)
	mflr    0
	stw     0,20(1)
	bl      get_onstack_sigmask
	lwz     0,20(1)
	mtlr    0
	addi    1,1,16

__obsolete_setup_setjmp:
	/* r3 = save area, r6 = onstack, r7 = sigmask */
	stwu    1,-80(1)
	mflr    0
	stmw    18,16(1)	/* save d2-a6, which are clobbered */
	stw     0,84(1)		/* by a call to EmulCallOS() in malloc() */
	stw     3,8(1)
	stw     6,0(3)
	stw     7,4(3)
	lmw     18,8(2)
	/* Technically this could be 8 bytes less or we could call memalign(x,16)
	   with 12 bytes less, but lets not get too smart - Piru */
	li      3,584
	bl      malloc
	cmpi    0,3,0
	lwz     4,8(1)
	bne     .Lok4
	b       abort
.Lok4:
	stmw    18,8(2)
	li      5,0
	lwz     0,84(1)
	stw     3,8(4)
	lmw     18,16(1)
	lwz     6,0(4)
	mtlr    0
	lwz     7,4(4)
	stw     5,24(4)
	addi    1,1,80
	b       setup_setjmp
");


ENTRY(_trampoline_sigsetjmp)
asm("
	lwz     5,60(2)
	lwz     4,8(5)
	lwz     3,4(5)
	b       sigsetjmp
");

ENTRY(sigsetjmp)
asm("
	cmpi    0,4,0
	stw     4,0(3)
	addi    3,3,4
	bne     setjmp
	b       _setjmp
");


ENTRY(_trampoline__setjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	li      6,0             /* dummy signal mask */
	li      7,0             /*  and onstack */
	b       setup_setjmp
");

ENTRY(_setjmp)
asm("
	li      6,0             /* dummy signal mask */
	li      7,0             /*  and onstack */
	b       setup_setjmp
");

ENTRY(_trampoline_setjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	b       setjmp
");

#define __MKS(x) #x
#define _MKS(x) __MKS(x)

#define save_uvmx(cr,pr,sg,ag) \
"	addi    " _MKS(ag) "," _MKS(ag) ",32\n" \
"	vperm   3," _MKS(pr) "," _MKS(cr) ",0\n" \
"	stvx    3,0," _MKS(sg) "\n"

#define save_avmx(cr,ag) \
"	stvx    " _MKS(cr) ",0," _MKS(ag) "\n" \
"	addi    " _MKS(ag) "," _MKS(ag) ",32\n"

ENTRY(setjmp)
asm(
"	stwu    1,-16(1)\n"
"	mflr    0\n"
"	stw     0,20(1)\n"
"	bl      get_onstack_sigmask\n"
"	lwz     0,20(1)\n"
"	addi    1,1,16\n"
"	mtlr    0\n"

"setup_setjmp:\n"
	/* r3 = save area, r6 = onstack, r7 = sigmask */
"	lis     4,SysBase@ha\n"
"	mr      8,1\n"             /* r8  = r1 */
"	lwz     4,SysBase@l(4)\n"
"	mr      9,2\n"             /* r9  = r2 */
"	lhz     10,0x126(4)\n"     /* r10 = TDNestCnt, IDNestCnt */
"	mfcr    12\n"              /* r12 = cr */
"	mflr    11\n"              /* r11 = lr */
"	stfd    14,104(3)\n"
"	stfd    15,112(3)\n"
"	ori     12,12,1\n"         /* make cr !=0 */
"	stfd    16,120(3)\n"
"	stfd    17,128(3)\n"
"	stfd    18,136(3)\n"
"	stfd    19,144(3)\n"
"	stmw    6,0(3)\n"
"	stfd    20,152(3)\n"
"	stfd    21,160(3)\n"
"	stfd    22,168(3)\n"
"	stfd    23,176(3)\n"
"	stfd    24,184(3)\n"
"	stfd    25,192(3)\n"
"	stfd    26,200(3)\n"
"	stfd    27,208(3)\n"
"	stfd    28,216(3)\n"
"	stfd    29,224(3)\n"
"	stfd    30,232(3)\n"
"	stfd    31,240(3)\n"
"	lmw     26,8(2)\n"         /* d2-d7 */
"	lwz     4,64(2)\n"         /* pc */
"	stmw    26,256(3)\n"
"	lmw     26,40(2)\n"        /* a2-a7 */
"	stw     4,248(3)\n"
"	lis     4,altivecflag@ha\n"
"	stmw    26,280(3)\n"
"	lwz     6,0(31)\n"         /* *a7 */
"	lwz     4,altivecflag@l(4)\n"
"	lmw     14,0x88(2)\n"      /* fp2-fp7 */
"	stw     6,252(3)\n"
"	mr.	4,4\n"
"	stmw    14,304(3)\n"

	/* Is altivec available? (altivecflag != 0) */
"	beq-    .Lsetjmp_done\n"

	/* Altivec is available, store vrsave and v20..v31.
	 * we need to be extra careful as the jmp_buf is only
	 * guaranteed to be aligned by 4. In practice we have
	 * to handle arbitrary alignment. Note that there are
	 * 12 extra bytes reserved after v31 to avoid race
	 * condition that could be possible if the next bytes
	 * could end up being shared with another task. In
	 * addition being race-safe, this also enables us to
	 * avoid having to deal with preserving the last bytes
	 * in case the vregs are not aligned. - Piru
	 */

"	mfvrsave 0\n"
"	addi    4,3,380\n"
"	stw     0,376(3)\n"   /* store vrsave */
"	andi.   5,4,15\n"
"	addi    5,4,16\n"
"	beq+    .Lsetjmp_aligned\n"

"	lvsr    0,0,4\n"      /* prepare vperm regs */
"	lvsl    1,0,4\n"
"	addi    5,4,-16\n"
"	lvx     2,0,4\n"      /* load & rotate memory before v20 */
"	vperm   2,2,2,1\n"

	save_uvmx(20,2,4,5)
	save_uvmx(21,20,5,4)
	save_uvmx(22,21,4,5)
	save_uvmx(23,22,5,4)
	save_uvmx(24,23,4,5)
	save_uvmx(25,24,5,4)
	save_uvmx(26,25,4,5)
	save_uvmx(27,26,5,4)
	save_uvmx(28,27,4,5)
	save_uvmx(29,28,5,4)
	save_uvmx(30,29,4,5)
	save_uvmx(31,30,5,4)
#if 1
	/* The 12 bytes are padding, so we don't need to care about
	 * their values and we can freely write trash there.
	 *
	 * The reason we have the padding rather than preserve the
	 * last 12 bytes, has to do with the nature of the MorphOS
	 * global memory model: If there was no padding, it could well
	 * happen that the next 12 bytes could belong to another task.
	 * If we would then be read-write modifying them, we might
	 * create a race condition if the app itself modifies the
	 * values between our accesses.
	 * - Piru
	 */
"	vperm   3,31,0,0\n"
"	stvx    3,0,4\n"
#else
"	lvx     2,0,5\n"      /* load & rotate memory after v31 */
"	vperm   2,2,2,1\n"
	save_uvmx(2,31,4,5)
#endif

"	b       .Lsetjmp_done\n"

".Lsetjmp_aligned:\n"
	save_avmx(20,4)
	save_avmx(21,5)
	save_avmx(22,4)
	save_avmx(23,5)
	save_avmx(24,4)
	save_avmx(25,5)
	save_avmx(26,4)
	save_avmx(27,5)
	save_avmx(28,4)
	save_avmx(29,5)
"	stvx    30,0,4\n"     /* v30 and v31 don't need increment */
"	stvx    31,0,5\n"

".Lsetjmp_done:\n"
"	lmw     14,32(3)\n"        /* restore r14-r31 */
"	li      3,0\n"             /* return 0 */
"	stw     3,0(2)\n"          /* in r3 and d0 */
"	blr\n"
);

/* _clean_longjmp, do NOT change onstack/sigmask, do NOT change stackframe */
/* this function is called directly by execve on_real_stack function */
ENTRY(_clean_longjmp)
asm("
	lwz     5,24(3)         /* old style buffer ? */
	cmpi    0,5,0
	bne+    .Lok5
	lwz     3,8(3)          /* yes -> get the new one */
.Lok5:
	stwu    1,-16(1)
	mflr    0
	stw     0,20(1)
	bl      get_onstack_sigmask
	lwz     0,20(1)
	mtlr    0
	addi    1,1,16

	lwz     5,8(3)          /* load sp */
	stw     6,0(3)          /* save current onstack value */
	cmpi    0,5,0           /* ensure non-zero SP */
	stw     7,4(3)          /* save current signal mask */
	cmpi    1,4,0           /* test return value */
	beq-    botch           /* bad stack ! */
	bne+    1,longjmp_cont  /* return value = non-zero ok */
	li      4,1             /* else make non-zero */
	b       longjmp_cont
");


ENTRY(_trampoline_siglongjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	lwz     4,8(5)
	b       siglongjmp
");

ENTRY(siglongjmp)
asm("
	lwz     5,24(3)         /* old style buffer ? */
	cmpi    0,5,0
	bne     .Lok6
	lwz     3,8(3)          /* yes -> get the new one */
.Lok6:
	lwz     5,0(3)
	addi    3,3,4
	cmpi    0,5,0
	bne     longjmp
	b       _longjmp
");

ENTRY(_trampoline__longjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	lwz     4,8(5)
	b       _longjmp
");

/* _longjmp, do NOT change onstack/sigmask */
ENTRY(_longjmp)
asm("
	lwz     5,24(3)         /* old style buffer ? */
	cmpi    0,5,0
	bne+    .Lok7
	lwz     3,8(3)          /* yes -> get the new one */
.Lok7:
	stwu    1,-16(1)
	mflr    0
	stw     0,20(1)
	bl      get_onstack_sigmask
	lwz     0,20(1)
	addi    1,1,16
	mtlr    0

	stw     6,0(3)          /* save current onstack value */
	stw     7,4(3)          /* save current signal mask */
	b       longjmp
");

ENTRY(_trampoline_longjmp)
asm("
	lwz     5,60(2)
	lwz     3,4(5)
	lwz     4,8(5)
	b       longjmp
");

#define load_uvmx(cr,pr,sg,ag) \
"	addi    " _MKS(ag) "," _MKS(ag) ",32\n" \
"	lvx     " _MKS(pr) ",0," _MKS(sg) "\n" \
"	vperm   " _MKS(cr) "," _MKS(cr) "," _MKS(pr) ",0\n"

#define load_avmx(cr,ag) \
"	lvx     " _MKS(cr) ",0," _MKS(ag) "\n" \
"	addi    " _MKS(ag) "," _MKS(ag) ",32\n"

ENTRY(longjmp)
asm(
"	lwz     5,24(3)\n"         /* old style buffer ? */
"	cmpi    0,5,0\n"
"	bne     .Lok8\n"
"	lwz     3,8(3)\n"          /* yes -> get the new one */
".Lok8:\n"
"	stwu	1,-16(1)\n"
"	stw	3,8(1)\n"
"	stw	4,12(1)\n"
"	lwz     3,300(3)\n"	/* r3 = new A7 */
"	lwz	4,60(2)\n"		/* r4 = old A7 */
"	bl	stkrst68k\n"	/* Go to the correct 68k stackframe */
"	lwz	3,8(1)\n"
"	lwz	4,12(1)\n"
"	addi	1,1,8\n"

"	lwz     5,8(3)\n"
"	cmpi    0,5,0\n"           /* ensure non-zero SP */
"	cmpi    1,4,0\n"           /* test return value */
"	beq-    botch\n"           /* oops! */
"	bne+    1,.Lok1\n"
"	li      4,1\n"             /* make return value non zero */
".Lok1:\n"
"	bl      __stkrst\n"        /* Go to correct stackframe, r5 should */
				/* contain the SP */

/* in:
 *   r3 - jmp_buf
 *   r4 - val
 */
"longjmp_cont:\n"
"	lis     5,altivecflag@ha\n"
"	lwz     2,12(3)\n"
"	lwz     5,altivecflag@l(5)\n"
"	lmw     14,304(3)\n"       /* fp2-fp7 */
"	mr.     5,5\n"
"	lwz     6,252(3)\n"        /* *a7 */
"	stmw    14,0x88(2)\n"
"	lmw     26,280(3)\n"       /* a2-a7 */
"	lwz     5,248(3)\n"        /* pc */
"	stmw    26,40(2)\n"
"	stw     6,0(31)\n"
"	lmw     26,256(3)\n"       /* d2-d7 */
"	stw     5,64(2)\n"
"	stmw    26,8(2)\n"
"	lmw     12,24(3)\n"
"	lfd     14,104(3)\n"
"	lfd     15,112(3)\n"
"	lfd     16,120(3)\n"
"	lfd     17,128(3)\n"
"	lfd     18,136(3)\n"
"	lfd     19,144(3)\n"
"	lfd     20,152(3)\n"
"	lfd     21,160(3)\n"
"	lfd     22,168(3)\n"
"	lfd     23,176(3)\n"
"	lfd     24,184(3)\n"
"	lfd     25,192(3)\n"
"	lfd     26,200(3)\n"
"	lfd     27,208(3)\n"
"	lfd     28,216(3)\n"
"	lfd     29,224(3)\n"
"	lfd     30,232(3)\n"
"	lfd     31,240(3)\n"

	/* Is altivec available? (altivecflag != 0) */
"	beq-    .Llongjmp_done\n"

"	addi    5,3,380\n"
"	lwz     0,376(3)\n"
"	andi.   6,5,15\n"
"	addi    6,5,16\n"
"	mtvrsave 0\n"         /* restore vrsave */
"	beq+    .Llongjmp_aligned\n"

"	lvsl    0,0,5\n"      /* prepare vperm regs */
"	lvx     1,0,5\n"
"	addi    5,5,32\n"

"	lvx     21,0,6\n"     /* handle v20 */
"	vperm   20,1,21,0\n"
	load_uvmx(21,22,5,6)
	load_uvmx(22,23,6,5)
	load_uvmx(23,24,5,6)
	load_uvmx(24,25,6,5)
	load_uvmx(25,26,5,6)
	load_uvmx(26,27,6,5)
	load_uvmx(27,28,5,6)
	load_uvmx(28,29,6,5)
	load_uvmx(29,30,5,6)
	load_uvmx(30,31,6,5)
"	lvx     1,0,5\n"      /* handle v31 */
"	vperm   31,31,1,0\n"
"	b       .Llongjmp_done\n"

".Llongjmp_aligned:\n"
	load_avmx(20,5)
	load_avmx(21,6)
	load_avmx(22,5)
	load_avmx(23,6)
	load_avmx(24,5)
	load_avmx(25,6)
	load_avmx(26,5)
	load_avmx(27,6)
	load_avmx(28,5)
	load_avmx(29,6)
"	lvx     30,0,5\n"     /* v30 and v31 don't need increment */
"	lvx     31,0,6\n"

".Llongjmp_done:\n"
"	mtcr    12\n"
"	b       sigreturn\n"

"botch:\n"
"	b       longjmperror\n"
);

#endif /* NATIVE_MORPHOS */
