/*      $NetBSD: funopen.c,v 1.4 1995/02/02 02:09:44 jtc Exp $  */

/*-
 * Copyright (c) 1990, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
#if 0
static char sccsid[] = "@(#)funopen.c   8.1 (Berkeley) 6/4/93";
#endif
static char rcsid[] = "$NetBSD: funopen.c,v 1.4 1995/02/02 02:09:44 jtc Exp $";
#endif /* LIBC_SCCS and not lint */

#define _KERNEL
#include "ixemul.h"

#include <stdio.h>
#include <errno.h>
#include "local.h"

#warning "Perhaps there should also be a funopen64 function?"

fpos64_t _call_seekfn(FILE *fp, fpos64_t off, int whence)
{
	usetup;

	/*
	 * Technically any caller of this function should have tested for
	 * this case already, but lets be paranoid.
	 */
	if (fp->_seek == 0)
	{
		errno = ESPIPE;
		return -1LL;
	}

	if (fp->_seek == (void *) funopen)
	{
		/*
		 * Ok this is a funopen special filehandle with 32bit 'fpos_t'
		 * offset. If the passed 'off' cannot fit into 32bit value, fail
		 * right away. Otherwise let the _seek32 determine the result,
		 * if available. - Piru
		 */
		if (fp->_seek32 == 0 || off < -0x80000000LL || off > 0x7fffffffLL)
		{
			errno = ESPIPE;
			return -1LL;
		}
		return (fpos64_t) fp->_seek32(fp->_cookie, (fpos_t) off, whence);
	}

	return fp->_seek(fp->_cookie, off, whence);
}

FILE *
funopen(cookie, readfn, writefn, seekfn, closefn)
	const void *cookie;
	int (*readfn)(), (*writefn)();
#if __STDC__
	fpos_t (*seekfn)(void *cookie, fpos_t off, int whence);
#else
	fpos_t (*seekfn)();
#endif
	int (*closefn)();
{
	register FILE *fp;
	int flags;
	usetup;

	if (readfn == NULL) {
		if (writefn == NULL) {          /* illegal */
			errno = EINVAL;
			return (NULL);
		} else
			flags = __SWR;          /* write only */
	} else {
		if (writefn == NULL)
			flags = __SRD;          /* read only */
		else
			flags = __SRW;          /* read-write */
	}
	if ((fp = __sfp()) == NULL)
		return (NULL);
	fp->_flags = flags;
	fp->_file = -1;
	fp->_cookie = (void *)cookie;
	fp->_read = readfn;
	fp->_write = writefn;
	fp->_seek = (void *) funopen; /* magic value meaning seek32 is to be used */
	fp->_seek32 = seekfn;
	fp->_close = closefn;
	return (fp);
}
