#define _KERNEL
#include "ixemul.h"

#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

static char *
__realpath(const char *path, char *resolved_path, int count)
{
	char buf[PATH_MAX + 1];
	char* c;
	int i;
	usetup;

	if(path == NULL)
	{
		errno = EINVAL;
		return 0;
	}

	if(path[0] == '\0')
	{
		errno = ENOENT;
		return 0;
	}

	if(count <= 0)
	{
		/* Too many symlinks ... */
		errno = ELOOP;
		return 0;
	}

	if(chdir(path) == 0)
		return getcwd(resolved_path, PATH_MAX);

	c = strrchr(path, '/');
	if(c)
	{
		if(c-path > PATH_MAX)
			return 0;

		memcpy(buf, path, c-path);
		buf[c-path] = '\0';
		if(chdir(buf) == -1)
			return 0;

		path = ++c;
	}

	if(readlink(path, buf, PATH_MAX) == 0)
		return __realpath(buf, resolved_path, --count);

	if(getcwd(resolved_path, PATH_MAX) == 0)
		return 0;
	i = strlen(resolved_path);
	resolved_path[i++] = '/';
	for(;i < PATH_MAX-1; ++i)
	{
		if (!(resolved_path[i] = *path))
			break;
		++path;
	}
	resolved_path[i] = '\0';
	return resolved_path;
}

char *
realpath(const char *path, char *resolved_path)
{
	char* res;
	int dofree = 0;
	int fd;
//	usetup;

	if (resolved_path == 0)
	{
		resolved_path = (char *) syscall(SYS_malloc, (u_int)PATH_MAX);
		if (resolved_path == 0)
		{
			return 0;
		}

		dofree = 1;
	}

	/* Save current dir */
	fd = syscall(SYS_open, ".", O_RDONLY);

	res = __realpath(path, resolved_path, SYMLOOP_MAX);

	/* Restore current dir */
	syscall(SYS_fchdir, fd);
	syscall(SYS_close, fd);

	if (res == 0 && dofree)
		syscall(SYS_free, resolved_path);

	return res;
}

