#define _LARGEFILE64_SOURCE 1
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#ifdef __linux__
#include <sys/vfs.h>
#endif
#include <fcntl.h>

#ifdef __libnix
const char *tname = "T:testfile.tmp";
const char *nname = "T:doesnotexist.tmp";
#else
const char *tname = "/tmp/testfile.tmp";
const char *nname = "/tmp/doesnotexist.tmp";
#endif
int tfd;
FILE *tfh;
int tsiz;

#define TBUFSIZ 65536
char tbuf[TBUFSIZ];

#ifndef __libnix
int statfs_validate(struct statfs *s32, struct statfs64 *s64)
{
  /* NOTE: Will bork with a large /tmp volume */
  if (s32->f_bsize != s64->f_bsize) return 0;
#ifndef __linux__
  if (s32->f_fsize != s64->f_fsize) return 0;
#endif
  if (s32->f_blocks != s64->f_blocks) return 0;
  if (s32->f_bfree != s64->f_bfree) return 0;
  if (s32->f_bavail != s64->f_bavail) return 0;
  if (s32->f_files != s64->f_files) return 0;
  if (s32->f_ffree != s64->f_ffree) return 0;
  return 1;
}
#endif

int t_getfsstat(void)
{
#if !defined(__linux__) && !defined(__libnix)
  struct statfs *s32 = (void *) tbuf;
  struct statfs64 *s64 = (void *) tbuf;
  int c32, c64, t;

  c32 = getfsstat(NULL, 0, 0);
  if (c32 == -1) return 0;
  c64 = getfsstat64(NULL, 0, 0);
  if (c64 == -1) return 0;
  if (c32 != c64) return 0;

  if (c32 != c64) return 0;
  memset(s32, 0xaa, TBUFSIZ);
  t = getfsstat(s32, TBUFSIZ, 0);
  if (t != c32) return 0;

  memset(s64, 0xaa, TBUFSIZ);
  t = getfsstat64(s64, TBUFSIZ, 0);
  if (t != c64) return 0;

#warning TODO: Add more validation
#endif

  return 1;
}

int t_fstatfs(void)
{
#ifndef __libnix
  struct statfs s32;
  struct statfs64 s64;

  if (statfs(nname, &s32) != -1) return 0;
  if (statfs64(nname, &s64) != -1) return 0;
  if (statfs(tname, &s32) != 0) return 0;
  if (statfs64(tname, &s64) != 0) return 0;

  return statfs_validate(&s32, &s64);
#else
  return 1;
#endif
}

int t_statfs(void)
{
#ifndef __libnix
  struct statfs s32;
  struct statfs64 s64;

  if (fstatfs(-1, &s32) != -1) return 0;
  if (fstatfs64(-1, &s64) != -1) return 0;
  if (fstatfs(555555, &s32) != -1) return 0;
  if (fstatfs64(555555, &s64) != -1) return 0;

  if (fstatfs(tfd, &s32) != 0) return 0;
  if (fstatfs64(tfd, &s64) != 0) return 0;

  return statfs_validate(&s32, &s64);
#else
  return 1;
#endif
}

int t_getmntinfo(void)
{
#if !defined(__linux__) && !defined(__libnix)
  struct statfs *s32;
  struct statfs64 *s64;
  int c32, c64, i;

  if ((c32 = getmntinfo(&s32, 0)) == 0) return 0;
  if ((c64 = getmntinfo64(&s64, 0)) == 0) return 0;
  if (c32 != c64) return 0;
  for (i = 0; i < c32; i++)
  {
    if (!strcmp(s32[i].f_mntfromname, "/Ram Disk")) continue;
    if (!statfs_validate(&s32[i], &s64[i])) return 0;
  }
#endif
  return 1;
}

int t_fseek_set(void)
{
  if (fseek(tfh, 0, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != 0) return 0;
  if (fseek(tfh, -1, SEEK_SET) != -1) return 0;
  if (ftell(tfh) != 0) return 0;
#ifdef __linux__
  if (fseek(tfh, tsiz + 1, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != tsiz + 1) return 0;
  if (fseek(tfh, 1000000000, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != 1000000000) return 0;
#endif

#ifndef __libnix
  if (fseeko64(tfh, 0, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != 0) return 0;
  if (fseeko64(tfh, -1, SEEK_SET) != -1) return 0;
  if (ftell(tfh) != 0) return 0;
#ifdef __linux__
  if (fseeko64(tfh, tsiz + 1, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != tsiz + 1) return 0;
  if (fseeko64(tfh, 100000000000LL, SEEK_SET) != 0) return 0;
  if (ftello64(tfh) != 100000000000LL) return 0;
#endif
#endif
  return 1;
}

int t_fseek_cur(void)
{
  if (fseek(tfh, 0, SEEK_SET) != 0) return 0;
  if (fseek(tfh, -1, SEEK_CUR) != -1) return 0;
  if (ftell(tfh) != 0) return 0;
  if (fseek(tfh, 1, SEEK_CUR) != 0) return 0;
  if (ftell(tfh) != 1) return 0;
#ifdef __linux__
  if (fseek(tfh, 10000, SEEK_CUR) != 0) return 0;
  if (ftell(tfh) != 10001) return 0;
  if (fseek(tfh, 1000010001, SEEK_SET) != 0) return 0;
  if (ftell(tfh) != 1000010001) return 0;
#endif

#ifndef __libnix
  if (fseek(tfh, 0, SEEK_SET) != 0) return 0;
  if (fseeko64(tfh, -1, SEEK_CUR) != -1) return 0;
  if (ftell(tfh) != 0) return 0;
  if (fseeko64(tfh, 1, SEEK_CUR) != 0) return 0;
  if (ftell(tfh) != 1) return 0;
#ifdef __linux__
  if (fseeko64(tfh, 10000, SEEK_CUR) != 0) return 0;
  if (ftell(tfh) != 10001) return 0;
  if (fseeko64(tfh, 100000010001LL, SEEK_SET) != 0) return 0;
  if (ftello64(tfh) != 100000010001ULL) return 0;
#endif
#endif
  return 1;
}

int t_fseek_end(void)
{
  if (fseek(tfh, 0, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz) return 0;
  if (fseek(tfh, -1, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz - 1) return 0;
#ifdef __linux__
  if (fseek(tfh, 1, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz + 1) return 0;
  if (fseek(tfh, 1000010001, SEEK_END) != 0) return 0;
  if (ftell(tfh) != (1000010001 + (uint32_t) tsiz)) return 0;
#endif

#ifndef __libnix
  if (fseeko64(tfh, 0, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz) return 0;
  if (fseeko64(tfh, -1, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz - 1) return 0;
#ifdef __linux__
  if (fseeko64(tfh, 1, SEEK_END) != 0) return 0;
  if (ftell(tfh) != tsiz + 1) return 0;
  if (fseeko64(tfh, 100000010001ULL, SEEK_END) != 0) return 0;
  if (ftello64(tfh) != (100000010001ULL + (uint64_t) tsiz)) return 0;
#endif
#endif
  return 1;
}

int t_fgetpos(void)
{
  fpos_t o32;
  fpos64_t o64;
  if (fseek(tfh, 12, SEEK_SET) != 0) return 0;
  if (fgetpos(tfh, &o32) != 0) return 0;
#ifdef __linux__
  if (o32.__pos != 12) return 0;
#else
  if (o32 != 12) return 0;
#endif

#ifdef __libnix
  if (fseek(tfh, 13, SEEK_SET) != 0) return 0;
#else
  if (fseeko64(tfh, 13LL, SEEK_SET) != 0) return 0;
#endif
  if (fgetpos64(tfh, &o64) != 0) return 0;
#ifdef __linux__
  if (o64.__pos != 13LL) return 0;
#else
#ifdef __libnix
  if (o64 != 13LL) return 0;
#else
  if (o64 != 13LL) return 0;
#endif
#endif

  return 1;
}

int t_fsetpos(void)
{
  fpos_t o32;
  fpos64_t o64;
  if (fseek(tfh, 12, SEEK_SET) != 0) return 0;
  if (fgetpos(tfh, &o32) != 0) return 0;
  if (fseek(tfh, 0, SEEK_SET) != 0) return 0;
  if (fsetpos(tfh, &o32) != 0) return 0;
  if (ftell(tfh) != 12) return 0;

#ifdef __libnix
  if (fseek(tfh, 13, SEEK_SET) != 0) return 0;
#else
  if (fseeko64(tfh, 13LL, SEEK_SET) != 0) return 0;
#endif
  if (fgetpos64(tfh, &o64) != 0) return 0;
#ifdef __libnix
  if (fseek(tfh, 0, SEEK_SET) != 0) return 0;
#else
  if (fseeko64(tfh, 0LL, SEEK_SET) != 0) return 0;
#endif
  if (fsetpos64(tfh, &o64) != 0) return 0;
#ifdef __libnix
  if (ftell(tfh) != 13) return 0;
#else
  if (ftello64(tfh) != 13LL) return 0;
#endif

  return 1;
}

int t_lseek(void)
{
  if (lseek(-1, 0, SEEK_SET) != -1) return 0;
  if (lseek(555555, 0, SEEK_SET) != -1) return 0;
  if (lseek(tfd, -1, SEEK_SET) != -1) return 0;
#ifdef __linux__
  if (lseek(tfd, 1000000, SEEK_SET) != 1000000) return 0;
  if (lseek(tfd, 0, SEEK_CUR) != 1000000) return 0;
#endif
  if (lseek(tfd, tsiz, SEEK_SET) != tsiz) return 0;
  if (lseek(tfd, -(tsiz+1), SEEK_CUR) != -1) return 0;
  if (lseek(tfd, 0, SEEK_END) != tsiz) return 0;
  if (lseek(tfd, -tsiz, SEEK_END) != 0) return 0;
#ifdef __linux__
  if (lseek(tfd, 1, SEEK_END) != tsiz + 1) return 0;
#endif
  if (lseek(tfd, -(tsiz+1), SEEK_END) != -1) return 0;

  if (lseek64(-1, 0LL, SEEK_SET) != -1LL) return 0;
  if (lseek64(555555, 0LL, SEEK_SET) != -1LL) return 0;
  if (lseek64(tfd, -1LL, SEEK_SET) != -1LL) return 0;
#ifdef __linux__
  if (lseek64(tfd, 100000000000LL, SEEK_SET) != 100000000000LL) return 0;
  if (lseek64(tfd, 0LL, SEEK_CUR) != 100000000000LL) return 0;
#endif
  if (lseek64(tfd, tsiz, SEEK_SET) != tsiz) return 0;
  if (lseek64(tfd, -(tsiz+1), SEEK_CUR) != -1LL) return 0;
  if (lseek64(tfd, 0LL, SEEK_END) != tsiz) return 0;
  if (lseek64(tfd, -tsiz, SEEK_END) != 0LL) return 0;
#ifdef __linux__
  if (lseek64(tfd, 1LL, SEEK_END) != tsiz + 1) return 0;
#endif
  if (lseek64(tfd, -(tsiz+1), SEEK_END) != -1LL) return 0;

  return 1;
}

int t_ftruncate(void)
{
  int fd = open(tname, O_RDONLY);
  if (fd == -1) return 0;

  if (lseek(tfd, 1000000, SEEK_SET) != 1000000) return 0;
  if (ftruncate(-1, 0) != -1) return 0;
  if (ftruncate(55555, 0) != -1) return 0;
  if (ftruncate(tfd, tsiz) != 0) { perror("ftruncate tsiz"); return 0; }
  if (lseek(tfd, 0, SEEK_CUR) != 1000000) { printf("WARNING: ftruncate seek position reset\n"); }
  if (ftruncate(fd, 0) != -1) { printf("could ftruncate O_RDONLY fd\n"); return 0; }

#ifdef __linux__
  if (lseek64(tfd, 100000000000LL, SEEK_SET) != 100000000000LL) return 0;
#else
  if (lseek64(tfd, 1000000LL, SEEK_SET) != 1000000LL) return 0;
#endif
  if (ftruncate64(-1, 0LL) != -1) return 0;
  if (ftruncate64(55555, 0LL) != -1) return 0;
  if (ftruncate64(tfd, tsiz) != 0) { perror("ftruncate64 tsiz"); return 0; }
#ifdef __linux__
  if (lseek64(tfd, 0LL, SEEK_CUR) != 100000000000LL) return 0;
#else
  if (lseek64(tfd, 0LL, SEEK_CUR) != 1000000LL) { printf("WARNING: ftruncate64 seek position reset\n"); }
#endif
  if (ftruncate64(fd, 0) != -1) { printf("could ftruncate64 O_RDONLY fd\n"); return 0; }

  close(fd);

  return 1;
}

int t_truncate(void)
{
  if (truncate(nname, 0) != -1) return 0;
  if (truncate(tname, tsiz) != 0) { perror("truncate tsiz"); return 0; }

  if (truncate64(nname, 0) != -1) return 0;
  if (truncate64(tname, tsiz) != 0) { perror("truncate64 tsiz"); return 0; }

  return 1;
}

int t_stat(void)
{
  struct stat s32;
  struct stat64 s64;

  if (stat(nname, &s32) != -1) return 0;
  if (lstat(nname, &s32) != -1) return 0;
  if (stat(tname, &s32) != 0) return 0;
  if (lstat(tname, &s32) != 0) return 0;
  if (s32.st_size != tsiz) return 0;

  if (stat64(nname, &s64) != -1) return 0;
  if (lstat64(nname, &s64) != -1) return 0;
  if (stat64(tname, &s64) != 0) return 0;
  if (lstat64(tname, &s64) != 0) return 0;
  if (s64.st_size != tsiz) return 0;
#ifdef __MORPHOS__
  if (s64.st_size32 != tsiz) return 0;
#endif

  return 1;
}

int t_fstat(void)
{
  struct stat s32;
  struct stat64 s64;

  if (fstat(-1, &s32) != -1) return 0;
  if (fstat(55555, &s32) != -1) return 0;
  if (fstat(tfd, &s32) != 0) return 0;
  if (s32.st_size != tsiz) return 0;

  if (fstat64(-1, &s64) != -1) return 0;
  if (fstat64(55555, &s64) != -1) return 0;
  if (fstat64(tfd, &s64) != 0) return 0;
  if (s64.st_size != tsiz) return 0;
#ifdef __MORPHOS__
  if (s64.st_size32 != tsiz) return 0;
#endif

  return 1;
}

int main(void)
{
  struct
  {
    const char *name;
    int (*func)(void);
  }
  tests [] =
  {
    {"getfsstat",   t_getfsstat},
    {"fstatfs",     t_fstatfs},
    {"statfs",      t_statfs},
    {"getmntinfo",  t_getmntinfo},
    {"fseek_set",   t_fseek_set},
    {"fseek_cur",   t_fseek_cur},
    {"fseek_end",   t_fseek_end},
    {"fgetpos",     t_fgetpos},
    {"fsetpos",     t_fsetpos},
    {"lseek",       t_lseek},
    {"truncate",    t_truncate},
    {"ftruncate",   t_ftruncate},
    {"stat",        t_stat},
    {"fstat",       t_fstat},
  };
  const int numtests = sizeof(tests) / sizeof(tests[0]);
  int i;
  struct stat st;

  tfh = fopen(tname, "w");
  if (!tfh) { perror("fopen"); return 0; }
  fprintf(tfh, "This is a test file\n");
  fclose(tfh);
  if (stat(tname, &st) == -1) { perror("stat"); return 0; }
  tsiz = st.st_size;

  for (i = 0; i < numtests; i++)
  {
    tfd = open(tname, O_RDWR);
    if (tfd == -1) { perror("open"); return 0; }
    tfh = fdopen(tfd, "r+");
    if (tfh == 0) { perror("fopen"); return 0; }

    printf("test %d: '%s' ", i, tests[i].name); fflush(stdout);
    printf("%s\n", tests[i].func() ? "OK" : "FAIL");

    fclose(tfh);
    close(tfd);
  }

  unlink(tname);

  return 1;
}
