#ifdef __AMIGA__
#include <proto/exec.h>
#ifndef __MORPHOS__
#define UTILITY_TAGITEM_H
#define _SIZE_T
#define __AMIGA_TYPES__
#endif

#include <devices/timer.h>
#include <ixemul.h>
#endif
#include <stdint.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <signal.h>
#include <arpa/inet.h>
#ifdef __AMIGA__
#include <ix.h>

char VERSION[] = "\000$VER: ixstack 1.3 (10.03.2020)";
#endif

#define STACK_MAGIC_ID0 0x5374436B /* 'StCk' */
#define STACK_MAGIC_ID1 0x7354634B /* 'sTcK' */

void setstack(int32_t size, char *filename)
{
  int fd;
  off_t len, i;
  caddr_t addr;
  struct stat s;
  static int printed_header = 0;
  
  if (stat(filename, &s) == -1 || !S_ISREG(s.st_mode))
  {
    printf("%s isn't a regular file\n", filename);
    return;
  }
  fd = open(filename, O_RDWR);
  if (fd == -1)
  {
    perror(filename);
    return;
  }
  len = lseek(fd, 0, SEEK_END);
  addr = mmap(NULL, len, PROT_READ, MAP_SHARED, fd, 0);
  if (fd == -1)
  {
    perror(filename);
    close(fd);
    return;
  }
  for (i = 0; i < len - 12; i += 2)
    if (*(uint32_t *)(addr + i) == htonl(STACK_MAGIC_ID0) &&
        *(uint32_t *)(addr + i + 8) == htonl(STACK_MAGIC_ID1))
    {
      if (size >= 0)
      {
	uint32_t s = htonl(size);
	if (lseek(fd, i + 4, SEEK_SET) != i + 4 ||
	    write(fd, &s, 4) != 4)
	  perror(filename);
      }
      else
      {
	if (!printed_header)
	{
	  printed_header = 1;
	  printf("stacksize  filename\n---------  --------\n");
	}
	printf("%9u  %s\n", ntohl(*(uint32_t *)(addr + i + 4)), filename);
      }
      munmap(addr, len);
      close(fd);
      return;
    }
  close(fd);
  if (size)
    printf("cannot %cet stack: %s\n", size >= 0 ? 's' : 'g', filename);
}

#ifdef __AMIGA__

static int ctrlc = 0;

static void sigint()
{
  ctrlc = 1;
}

static void show(void)
{
  struct MsgPort *port;
  struct SUMessage *msg;
  ULONG portsig;
  int printed_header = 0;
  const char portname[] = "ixstack port";
  int msgtype = -1;

  /* Only allow SIGINT */
  int oldmask = sigsetmask(~sigmask(SIGINT));
  signal(SIGINT, sigint);

  /* Allow only one incarnation */
  Forbid();
  port = FindPort(portname);
  if (port)
  {
    Permit();
    sigsetmask(oldmask);
    fprintf(stderr, "ixstack show already running elsewhere\n");
    exit(1);
  }
  port = CreateMsgPort();
  if (port)
  {
    port->mp_Node.ln_Name = (STRPTR) portname;
    port->mp_Node.ln_Pri  = 0;
    AddPort(port);
  }
  Permit();

  if (port)
  {
    while (!ctrlc)
    {
      portsig = 1 << port->mp_SigBit | SIGBREAKF_CTRL_C;
      ix_wait((long *) &portsig);
      if (portsig & (1 << port->mp_SigBit))
      {
	while ((msg = (struct SUMessage *) GetMsg(port)))
	{
	  if (msgtype == -1)
	  {
	    switch (msg->msg.mn_Length)
	    {
	      case sizeof(*msg):
		msgtype = 1;
		break;
	      default:
	      /*case offsetof(struct SUMessage, ppc_stack_usage):*/
		msgtype = 0;
		break;
            }
	  }

	  if (!printed_header)
	  {
	    printed_header = 1;
	    if (msgtype == 1)
	      printf("  Usage   Total 68KUsage 68KTotal Program\n\n");
	    else
	      printf("  Usage   Total Program\n\n");
	  }
	  if (msgtype == 1)
	    printf("%7lu %7lu %8lu %8lu %s\n", msg->ppc_stack_usage, msg->ppc_stack_size, msg->stack_usage, msg->stack_size, msg->name);
	  else
	    printf("%7lu %7lu %s\n", msg->stack_usage, msg->stack_size, msg->name);
	  ReplyMsg(&msg->msg);
	}
      }
      
      if (portsig & SIGBREAKF_CTRL_C)
      {
	break;
      }
    }

    RemPort(port);

    /* Flush any pending messages */
    while ((msg = (struct SUMessage *) GetMsg(port)))
      ReplyMsg(&msg->msg);

    DeleteMsgPort(port);
  }

  sigsetmask(oldmask);

  exit(0);
}
#endif

int main(int argc, char **argv)
{
  int32_t size;

#ifdef __AMIGA__
  if (argc == 2 && !strcmp(argv[1], "-s"))
    show();
#endif
  if (argc < 3)
  {
#ifdef __AMIGA__
    fprintf(stderr, "set stacksize:   ixstack <stacksize> <files ...>\n"
                    "show stacksize:  ixstack -l <files ...>\n"
                    "show stackusage: ixstack -s\n");
#else
    fprintf(stderr, "set stacksize:   ixstack <stacksize> <files ...>\n"
                    "show stacksize:  ixstack -l <files ...>\n");
#endif
    exit(1);
  }

  if (!strcmp(argv[1], "-l"))
    size = -1;
  else
  {
    int i;
    
    for (i = 0; argv[1][i]; i++)
      if (!isdigit(argv[1][i]))
      {
	fprintf(stderr, "stacksize %s is not a number\n", argv[1]);
	exit(1);
      }
    size = atol(argv[1]);
    if (size < 4000 && size)
    {
      fprintf(stderr, "stacksize must be at least 4000 bytes\n");
      exit(1);
    }
    if (size && (size & 3))
    {
      fprintf(stderr, "stacksize must be a multiple of 4\n");
      exit(1);
    }
  }
  argv += 2;
  while (*argv)
    setstack(size, *argv++);
  return 0;
}
