#define _KERNEL
#include "ixemul.h"
#include <wchar.h>
#include <errno.h>

size_t mbrtowc(wchar_t *pwc, const char *s, size_t n, mbstate_t *ps)
{
	usetup;
	size_t i;

	if(!ps)
	{
		ps = &u_ptr->u_is;
	}

	if(!s)
	{
		if(ps->count)
		{
			errno = EILSEQ;
			return (size_t) - 1;
		}
		else
		{
			ps->count = ps->sofar = 0;
			return 0;
		}
	}

	for(i = 0; i < n; ++i)
	{
		unsigned char c = s[i];

		if(ps->count)
		{
			/* we have an unfinished multibyte sequence */
			if((c & 0xc0) != 0x80)
			{
				errno = EILSEQ;
				ps->count = 0;
				return (size_t) -1;
			}

			ps->sofar = (ps->sofar << 6) + (c & 0x3f);

			if(!--ps->count)
			{
				if(pwc)
				{
					*pwc = ps->sofar;
					++pwc;
				}

				if (ps->sofar)
				{
					ps->sofar = 0;
					return i + 1;
				}
				else
				{
					ps->count = ps->sofar = 0;
					return 0;
				}
			}
		}
		else
		{
			if(c & 0x80)
			{
				/* multibyte ? */
				unsigned char x = c << 1;
				unsigned char cnt = 0;

				while(x & 0x80)
				{
					x <<= 1;
					++cnt;
				}

				if(!cnt || cnt > 5)
				{
					errno = EILSEQ;
					ps->count = 0;
					return (size_t) -1;
				}

				ps->sofar = x >> (cnt + 1);
				ps->count = cnt;
			}
			else
			{
				ps->sofar = c;

				if(pwc)
				{
					*pwc = ps->sofar;
					++pwc;
				}

				if(ps->sofar)
				{
					ps->sofar = 0;
					return i+1;
				}
				else
				{
					ps->count = ps->sofar = 0;
					return 0;
				}
			}
		}
	}

	return n;
}
