#define _KERNEL
#include "ixemul.h"
#include <wctype.h>
#include <ctype.h>

#undef iswalnum
int iswalnum(wint_t c)
{
	return c <= 255 ? isalnum(c) : 0;
}

#undef iswalpha
int iswalpha(wint_t c)
{
	return c <= 255 ? isalpha(c) : 0;
}

#undef iswblank
int iswblank(wint_t c)
{
	return c <= 255 ? isblank(c) : 0;
}

#undef iswcntrl
int iswcntrl(wint_t c)
{
	return c <= 255 ? iscntrl(c) : 0;
}

#undef iswdigit
int iswdigit(wint_t c)
{
	return c <= 255 ? isdigit(c) : 0;
}

#undef iswgraph
int iswgraph(wint_t c)
{
	return c <= 255 ? isgraph(c) : 0;
}

#undef iswlower
int iswlower(wint_t c)
{
	return c <= 255 ? islower(c) : 0;
}

#undef iswprint
int iswprint(wint_t c)
{
	return c <= 255 ? isprint(c) : 0;
}

#undef iswpunct
int iswpunct(wint_t c)
{
	return c <= 255 ? ispunct(c) : 0;
}

#undef iswspace
int iswspace(wint_t c)
{
	return c <= 255 ? isspace(c) : 0;
}

#undef iswupper
int iswupper(wint_t c)
{
	return c <= 255 ? isupper(c) : 0;
}

#undef iswxdigit
int iswxdigit(wint_t c)
{
	return c <= 255 ? isxdigit(c) : 0;
}

#undef towlower
wint_t towlower(wint_t c)
{
	return c <= 255 ? tolower(c) : c;
}

#undef towupper
wint_t towupper(wint_t c)
{
	return c <= 255 ? toupper(c) : c;
}

wctype_t wctype(const char *s)
{
	static const char * const names[] =
	{
		"alnum",
		"alpha",
		"blank",
		"cntrl",
		"digit",
		"graph",
		"lower",
		"print",
		"punct",
		"space",
		"upper",
		"xdigit",
		NULL
	};

	const char * const *p;
	unsigned n = 1;

	for (p = names; *p; ++p, ++n)
	{
		if (strcmp(s, *p) == 0)
		{
			return n;
		}
	}

	return 0;
}

int iswctype(wint_t c, wctype_t d)
{
	int (* const func[])(wint_t) =
	{
		iswalnum,
		iswalpha,
		iswblank,
		iswcntrl,
		iswdigit,
		iswgraph,
		iswlower,
		iswprint,
		iswpunct,
		iswspace,
		iswupper,
		iswxdigit
	};
	const int numfunc = sizeof(func) / sizeof(func[0]);

	int r = 0;

	if (d >= 1 && d <= numfunc)
		r = func[d - 1](c);

	return r;
}

